/*
 * ASTROADD is distributed under the GNU GENERAL PUBLIC LICENSE (GPL)
 * Version 2 (June 1991). See the "COPYING" file distributed with this
 * software for more info.
 *
 * Copyright (C) 2000 by Michael Schroeder <mdhs@gmx.net>
 */


/*
 * Um aufwendige Neuberechnungen zwischen Funktionsaufrufen zu minimieren,
 * werden berechnete Werte in statischen Variablen zwischengespeichert und
 * bei erneutem Aufruf wieder ausgelesen (wenn gleiche Argumente vorliegen).
 *
 * WICHTIG: Das Eingabe-Datum ist StarCalc-basiert: 0 = 1899 Dec 30
 */


#define INCL_NLANGUAGE	1
//#define INCL_SYSINFO	1

#include <locale.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>

#include <fmath.h>
#include <fmathctl.h>
#include <fastmath.h>
#include <astro.h>

#include "starcalc.h"

#ifdef INCL_SYSINFO
# include <fcntl.h>
# include <netdb.h>
# include <pwd.h>
# include <unistd.h>

# include <sys/types.h>
# include <sys/stat.h>
# include <sys/utsname.h>
# include <sys/sysinfo.h>

# define IP(x,i)	((unsigned char)(x->h_addr[i]))
#endif

#define SC_COPYRIGHT	SC_copyright
#define SC_SYSINFO	SC_sysinfo
#define SC_UPDATE	SC_update
#define SC_DATE2	SC_date2
#define SC_EASTER	SC_easter
#define SC_DMS		SC_dms
#define SC_HMS		SC_hms
#define SC_DATIME	SC_datime
#define SC_OBSERVER	SC_observer
#define SC_TIMEZONE	SC_timezone
#define SC_POSITION	SC_position
#define SC_VISIBLE	SC_visible
#define SC_PHYSICAL	SC_physical
#define SC_ROTATION	SC_rotation
#define SC_BESSELDN	SC_besseldn
#define SC_LUNATION	SC_lunation
#define SC_MOONPHASE	SC_moonphase
#define SC_APSIDE	SC_apside
#define SC_MOONAPSIDE	SC_moonapside
#define SC_NODE		SC_node
#define SC_MOONNODE	SC_moonnode
#define SC_TROPIC	SC_tropic
#define SC_MOONMAXDECL	SC_moonmaxdecl
#define SC_NDECL	SC_ndecl
#define SC_STIME	SC_stime
#define SC_SDECL	SC_sdecl
#define SC_ECLIPSE	SC_eclipse
#define SC_ECLTYPE	SC_ecltype
#define SC_SEASON	SC_season

/*
 * Zeichen zu einer Nummer zusammenbasteln (little endian!)
 */
#define MKNUM(a,b)	((((unsigned char)b)<<8)|((unsigned char)a))

/*
 * Zeichen ausblenden und in Grossbuchstaben umwandeln (auch ANSI-Umlaute)
 */
#define GET1(c)		(*(unsigned char*)(c) & 0xdf)
#define GET2(c)		(*(unsigned short*)(c) & 0xdfdf)

#define UINIT		(1e300)			/* "noch unbelegt" */
#define INVALID		F_QNAN			/* Err:503 (ungueltig) */

extern char *_tscan(double *, char *);


typedef enum
{
  FUNC_DATE2 = 0,
#ifdef HAVE_UPDATE
  FUNC_UPDATE,
#endif
#ifdef INCL_SYSINFO
  FUNC_SYSINFO,
#endif

  FUNC_EASTER,
  FUNC_DMS,
  FUNC_HMS,
  FUNC_DATIME,
  FUNC_TIMEZONE,

  FUNC_POSITION,
  FUNC_VISIBLE,
  FUNC_PHYSICAL,
  FUNC_ROTATION,
  FUNC_BESSELDN,

  FUNC_LUNATION,
  FUNC_MOONPHASE,
  FUNC_APSIDE,
  FUNC_MOONAPSIDE,
  FUNC_NODE,
  FUNC_MOONNODE,
  FUNC_TROPIC,
  FUNC_MOONMAXDECL,
  FUNC_ECLIPSE,
  FUNC_ECLTYPE,
  FUNC_SEASON,
  COUNT_OF_FUNCS,	/* dieser Wert ist die Anzahl aller Funktionen */ 
} FUNCS;


typedef enum
{
#ifdef HAVE_UPDATE
  UPDATE_NAME,
  UPDATE_DESC,
#endif
#ifdef INCL_SYSINFO
  SYSINFO_NAME,
  SYSINFO_DESC,
  SYSINFOS_DESC,
#endif
  NUMBER_NAME,
  NUMBER_DESC,
  STRING_NAME,
  STRING_DESC,
  TAB_NAME,
  TAB_DESC,
  UNIT_NAME,
  UNIT_DESC,
  DATE_NAME,
  DATE_DESC,
  DAFMT_NAME,
  DAFMT_DESC,
  YEAR_NAME,
  YEAR_DESC,

  BODY_NAME,
  BODY_DESC,

  DATE2_NAME,
  DATE2_DESC,
  EASTER_NAME,
  EASTER_DESC,
  DMS_NAME,
  DMS_DESC,
  HMS_NAME,
  HMS_DESC,
  DATIME_NAME,
  DATIME_DESC,

  LONGITUDE_NAME,
  LONGITUDE_DESC,
  LATITUDE_NAME,
  LATITUDE_DESC,

  TIMEZONE_NAME,
  TIMEZONE_DESC,
  TZONES_DESC,

  POSITION_NAME,
  POSITION_DESC,
  POSITIONS_DESC,
  CENTER_NAME,
  CENTER_DESC,
  PLANE_NAME,
  PLANE_DESC,
  TYPE_NAME,
  TYPE_DESC,

  EQUINOX_NAME,
  EQUINOX_DESC,

  VISIBLE_NAME,
  VISIBLE_DESC,
  VISIBLES_DESC,

  PHYSICAL_NAME,
  PHYSICAL_DESC,
  PHYSICALS_DESC,

  ROTATION_NAME,
  ROTATION_DESC,
  ROTATIONS_DESC,

  BESSELDN_NAME,
  BESSELDN_DESC,
  BESSELDNS_DESC,

  LUNATION_NAME,
  LUNATION_DESC,
  MOONPHASE_NAME,
  MOONPHASE_DESC,
  LUNARPHASES_DESC,

  APSIDE_NAME,
  APSIDE_DESC,
  MOONAPSIDE_NAME,
  MOONAPSIDE_DESC,
  LUNARAPSIDES_DESC,

  NODE_NAME,
  NODE_DESC,
  MOONNODE_NAME,
  MOONNODE_DESC,
  LUNARNODES_DESC,

  TROPIC_NAME,
  TROPIC_DESC,
  MOONMAXDECL_NAME,
  MOONMAXDECL_DESC,
  LUNARMAXDECLS_DESC,

  ECLIPSE_NAME,
  ECLIPSE_DESC,
  ECLIPSES_DESC,
  ECLVALUE_DESC,
  ECLTYPE_DESC,
  ECLTYPE_NAME,
  ECLTYPES_DESC,

  SEASON_NAME,
  SEASON_DESC,
  SEASONS_DESC,
} STUFF;


typedef enum
{
  ECL_TYPE,
  ECL_TIME,
  ECL_MAX,
  ECL_TOTAL,
  ECL_PARTIAL,
  ECL_PENUMBRAL,
  ECL_G,
  ECL_U,
  ECL_LUNATION,
} ECLSTUFF;


/*
 * Statische Variable
 */
static const char *atname;	/* Kuerzel der aktuellen Zeitzone */
static double atzone, stzone;	/* Zeitdifferenzen zu UT */
static TRANS trans;		/* Weichensteller */


/*
 * Hier folgen kurze Erlaeuterungen zu den Funktionen und Argumenten
 * in verschiedenen Sprachen (auf richtigen Zeichensatz achten) 
 */
static char *getText(int nResource)
{
  switch(nResource) {
#ifdef HAVE_UPDATE
    EXPLAIN(UPDATE_NAME,
	    "update",
	    "Update");
    EXPLAIN(UPDATE_DESC,
	    "The function returns allways 0. It is used to update a cell"
	    " value immediately",
	    "Die Funktion liefert immer 0. Sie wird benutzt um einen Zell"
	    "wert sofort zu aktualisieren");
#endif
#ifdef INCL_SYSINFO
    EXPLAIN(SYSINFO_NAME,
	    "sysinfo",
	    "Sysinfo");
    EXPLAIN(SYSINFO_DESC,
	    "Gets general system informations",
	    "Gibt allgemeine Informationen zum System aus");
 /* EXPLAIN(STRING_NAME... */
    EXPLAIN(SYSINFOS_DESC,
	    "Possible values: user, home, host, system, release, machine"
	    ", FPU, path, RAM, memory, shared,... , /... (read file)",
	    "M"oe"gliche Werte: User, Home, Host, System, Release, Machine,"
	    " FPU, Path, RAM, Memory, Shared,... , /... (Datei lesen)");
#endif
    EXPLAIN(NUMBER_NAME,
	    "value",
	    "Zahl");
    EXPLAIN(NUMBER_DESC,
	    "value",
	    "Zahl");
    EXPLAIN(STRING_NAME,
	    "subfunction",
	    "Unterfunktion");
    EXPLAIN(STRING_DESC,
	    "name of the requested subfunction",
	    "Name der gew"ue"nschten Unterfunktion");

    EXPLAIN(TAB_NAME,
	    "fractional part",
	    "Nachkommastellen");
    EXPLAIN(TAB_DESC,
	    "Sets the number of digits of the fractional part",
	    "Legt die Zahl der angezeigten Nachkommastellen fest");

    EXPLAIN(UNIT_NAME,
	    "unit chars",
	    "Einheitenzeichen");
    EXPLAIN(UNIT_DESC,
	    "Displays the characters as units. An empty String forces use of"
	    " builtin units",
	    "Zeigt die Zeichen als Einheiten an. Eine leere Zeichenkette"
	    " veranlasst Verwendung der internen Einheitenzeichen");
    EXPLAIN(DAFMT_NAME,
	    "formatstring",
	    "Formatstring");
    EXPLAIN(DAFMT_DESC,
	    "Displays the date controlled by the format string:"
	    " [D][Y][M|M$][D|D#][Y][H][M][S|S#]",
	    "Zeigt das Datum nach den Regeln des Formatstrings an:"
	    " [T][J][M|M$][T|T#][J][H][M][S|S#]");

    EXPLAIN(DATE_NAME,
	    "date",
	    "Datum");
    EXPLAIN(DATE_DESC,
	    "date (extended range)",
	    "Datum (erweiterter Bereich)");
    EXPLAIN(YEAR_NAME,
	    "year",
	    "Jahr");
    EXPLAIN(YEAR_DESC,
	    "number of the year",
	    "Jahreszahl");

    EXPLAIN(DATE2_NAME,
	    "date2",
	    "Datum2");
    EXPLAIN(DATE2_DESC,
	    "Converts any date to intern number. (Is not subject to "
	    " StarOffice limits)",
	    "Liefert die interne Zahl f"ue"r ein beliebiges Datum. (Unterliegt"
	    " nicht den StarOffice Beschr"ae"nkungen)");

    EXPLAIN(EASTER_NAME,
	    "easter",
	    "Ostern");
    EXPLAIN(EASTER_DESC,
	    "Compute the date of Easter (Is not subject to StarOffice limits)",
	    "Berechnet den Ostertermin (Unterliegt nicht den StarOffice"
	    " Beschr"ae"nkungen)");

    EXPLAIN(DMS_NAME,
	    "DMS",
	    "DMS");
    EXPLAIN(DMS_DESC,
	    "Converts a decimal angle to an angle string (degrees, minutes,"
	    " seconds)",
	    "Wandelt einen dezimalen Winkel in eine sexagesimale Zeichenkette"
	    " um (Grad, Minuten, Sekunden)");
    EXPLAIN(HMS_NAME,
	    "HMS",
	    "HMS");
    EXPLAIN(HMS_DESC,
	    "Converts a decimal angle to an angle string (hours, minutes,"
	    " seconds)",
	    "Wandelt einen dezimalen Winkel in eine sexagesimale Zeichenkette"
	    " um (Stunden, Minuten, Sekunden)");
    EXPLAIN(DATIME_NAME,
	    "formatdate",
	    "FORMATZEIT");
    EXPLAIN(DATIME_DESC,
	    "Formats date and time. (Is not subject to StarOffice limits)",
	    "Formatiert Datum und Zeit (Unterliegt nicht den StarOffice"
	    " Beschr"ae"nkungen)");

    EXPLAIN(LONGITUDE_NAME,
	    "longitude",
	    "L"ae"nge");
    EXPLAIN(LONGITUDE_DESC,
	    "Sets the longitude",
	    "Stellt die geografische L"ae"nge ein");
    EXPLAIN(LATITUDE_NAME,
	    "latitude",
	    "Breite");
    EXPLAIN(LATITUDE_DESC,
	    "Sets the latitude",
	    "Stellt die geografische Breite ein");

    EXPLAIN(TIMEZONE_NAME,
	    "timezone",
	    "Zeitzone");
    EXPLAIN(TIMEZONE_DESC,
	    "Sets the timezone",
	    "Stellt die Zeitzone ein");
 /* EXPLAIN(STRING_NAME... */
    EXPLAIN(TZONES_DESC,
	    "Possible values: actual, system (without dst), TD, UT, none"
	    " HHH$$$ (time & text = zone time and name)",
	    "M"oe"gliche Werte: Aktuell, System (ohne Sommerzeit), TD, UT, Keine,"
	    " HHH$$$ (Zeit & Text = Zonenzeit- und Name)");

    EXPLAIN(BODY_NAME,
	    "object",
	    "Objekt");
    EXPLAIN(BODY_DESC,
	    "Possible values: Sun, Mercury, Venus, Mars, Jupiter,"
	    " Saturn, Uranus, Neptune, Pluto, Moon",
	    "M"oe"gliche Werte: Sonne, Merkur, Venus, Mars, Jupiter,"
	    " Saturn, Uranus, Neptun, Pluto, Mond");

    EXPLAIN(POSITION_NAME,
	    "position",
	    "Position");
    EXPLAIN(POSITION_DESC,
	    "Calculates the coordinates of the celestial body",
	    "Berechnet die Koordinaten des Himmelsk"oe"rpers");
 /* EXPLAIN(STRING_NAME... */
    EXPLAIN(POSITIONS_DESC,
	    "Possible values: r (g), l, b, x, y, z",
	    "M"oe"gliche Werte: r (w), l, b, x, y, z");

    EXPLAIN(CENTER_NAME,
	    "center of coordinates",
	    "Koordinaten-Ursprung");
    EXPLAIN(CENTER_DESC,
	    "Possible values: heliocentric, geocentric, topocentric",
	    "M"oe"gliche Werte: heliozentrisch, geozentrisch, topozentrisch");
    EXPLAIN(PLANE_NAME,
	    "reference plane",
	    "Bezugsebene");
    EXPLAIN(PLANE_DESC,
	    "Possible values: ecliptical, equatorial, horizontal",
	    "M"oe"gliche Werte: ekliptikal, "ae"quatorial, horizontal");
    EXPLAIN(TYPE_NAME,
	    "type",
	    "Typus");
    EXPLAIN(TYPE_DESC,
	    "Possible values: mean, true, apparent",
	    "M"oe"gliche Werte: mittel, wahr, scheinbar");
    EXPLAIN(EQUINOX_NAME,
	    "equinox",
	    AE "quinoktium");
    EXPLAIN(EQUINOX_DESC,
	    "Possible values: date, J2000, B1950, ...",
	    "M"oe"gliche Werte: Datum, J2000, B1950, ...");

    EXPLAIN(VISIBLE_NAME,
	    "visible",
	    "Sichtbar");
    EXPLAIN(VISIBLE_DESC,
	    "Calculates the times of rising, transit, setting and twilight",
	    "Berechnet die Aufg"ae"nge, Durchg"ae"nge, Unterg"ae"nge und D"ae"mmerungen");
 /* EXPLAIN(STRING_NAME... */
    EXPLAIN(VISIBLES_DESC,
	    "Possible values: rise, transit, set, culmination"
	    " cb, ce, nb, ne, ab, ae",
	    "M"oe"gliche Werte: Aufgang, Durchgang, Untergang, Kulmination,"
	    " BB, BE, NB, NE, AB, AE");

    EXPLAIN(PHYSICAL_NAME,
	    "physical",
	    "Physisch");
    EXPLAIN(PHYSICAL_DESC,
	    "Calculates the ephemeris for physical observations",
	    "Berechnet die physischen Ephemeriden");
 /* EXPLAIN(STRING_NAME... */
    EXPLAIN(PHYSICALS_DESC,
	    "Possible values: defect (q), elongation, magnitude, lighttime "
	    " (lt), parallax, phase (k), i, pradius, positionangle, radius",
	    "M"oe"gliche Werte: Defekt (q), Elongation, Helligkeit (m), Lichtzeit"
	    " (LZ), Parallaxe, Phase (k), i, PRadius, Positionswinkel, Radius");

    EXPLAIN(ROTATION_NAME,
	    "roatation",
	    "Rotation");
    EXPLAIN(ROTATION_DESC,
	    "Calculates the cartographic coordinates for the celestial body",
	    "Berechnet die kartografischen Koordinaten f"ue"r den Himmelsk"ae"rper");
 /* EXPLAIN(STRING_NAME... */
    EXPLAIN(ROTATIONS_DESC,
	    "Possible values: pos, l, l1, l2, l3, b, beta,"
	    " colongitude, terminator",
	    "M"oe"gliche Werte: pos, l, l1, l2, l3, b, beta,"
	    " Colongitude, Lichtgrenze");

    EXPLAIN(BESSELDN_NAME,
	    "besseldaynum",
	    "Besseltagzahl");
    EXPLAIN(BESSELDN_DESC,
	    "Calculates the Besselian day numbers",
	    "Berechnet die Besselschen Tageszahlen");
 /* EXPLAIN(STRING_NAME... */
    EXPLAIN(BESSELDNS_DESC,
	    "Possible values: A, B, C, D, E, mn, taneps, obliquity, X, Y",
	    "M"oe"gliche Werte: A, B, C, D, E, mn, taneps, Schiefe, X, Y");

    EXPLAIN(LUNATION_NAME,
	    "lunation",
	    "Lunation");
    EXPLAIN(LUNATION_DESC,
	    "Lunation is the number of conjunctions of the Moon and the Sun"
	    " i.e. the synodic months (New Moons) since 1/1/2000",
	    "Lunation ist die Zahl der Konjunktionen des Mondes mit der Sonne,"
	    " d.h. die synodischen Monate (Neumonde) seit dem 1.1.2000");
    EXPLAIN(MOONPHASE_NAME,
	    "moonphase",
	    "Mondphase");
    EXPLAIN(MOONPHASE_DESC,
	    "Computes the time of the phases of the Moon",
	    "Berechnet die Zeitpunkte der Mondphasen");
 /* EXPLAIN(STRING_NAME... */
    EXPLAIN(LUNARPHASES_DESC,
	    "possible values: newmoon, firstquater, fullmoon, lastquater",
	    "M"oe"gliche Werte: Neumond, Erstesviertel, Vollmond, Letztesviertel");

    EXPLAIN(APSIDE_NAME,
	    "apside",
	    "Apside");
    EXPLAIN(APSIDE_DESC,
	    "Apside is the number of conjunctions of the Moon and the"
	    " perigee, i.e. the anomalistic months since 1/1/2000",
	    "Apside ist die Zahl der Konjunktionen des Mondes mit dem"
	    " Perig"ae"um, d.h. die anomalistischen Monate seit dem 1.1.2000");
    EXPLAIN(MOONAPSIDE_NAME,
	    "moonapside",
	    "Mondapside");
    EXPLAIN(MOONAPSIDE_DESC,
	    "Computes the apsides of the Moon",
	    "Berechnet die Apsiden des Mondes");
 /* EXPLAIN(STRING_NAME... */
    EXPLAIN(LUNARAPSIDES_DESC,
	    "Possible values: perigee, ptime, apogee, atime",
	    "M"ae"gliche Werte: Perig"ae"um, PZeit, Apog"ae"um, AZeit");

    EXPLAIN(NODE_NAME,
	    "node",
	    "Knoten");
    EXPLAIN(NODE_DESC,
	    "Node is the number of conjunctions of the Moon and the"
	    " ascending node, i.e. the draconitic months since 1/1/2000",
	    "Knoten ist die Zahl der Konjunktionen des Mondes mit dem"
	    " aufsteigenden Knoten, d.h. die drakonitischen Monate seit"
	    " dem 1.1.2000");
    EXPLAIN(MOONNODE_NAME,
	    "moonnode",
	    "Mondknoten");
    EXPLAIN(MOONNODE_DESC,
	    "Computes the times of the passages of the Moon through the nodes",
	    "Berechnet die Zeitpunkte der Knotendurchg"ae"nge des Mondes");
 /* EXPLAIN(STRING_NAME... */
    EXPLAIN(LUNARNODES_DESC,
	    "Possible values: ascending, descending",
	    "M"oe"gliche Werte: Aufsteigend, Absteigend");

    EXPLAIN(TROPIC_NAME,
	    "tropic",
	    "Trope");
    EXPLAIN(TROPIC_DESC,
	    "Tropic is the number of conjunctions of the Moon and the"
	    " equinox, i.e. the tropic months since 1/1/2000",
	    "Trope ist die Zahl der Konjunktionen des Mondes mit dem"
	    " Fr"ue"hlingspunkt, d.h. die tropischen Monate seit dem 1.1.2000");
    EXPLAIN(MOONMAXDECL_NAME,
	    "moonmaxdecl",
	    "Mondmaxdekl");
    EXPLAIN(MOONMAXDECL_DESC,
	    "Computes the maximum declinations of the Moon",
	    "Berechnet die maximalen Deklinationen des Mondes");
 /* EXPLAIN(STRING_NAME... */
    EXPLAIN(LUNARMAXDECLS_DESC,
	    "Possible values: north, ntime, south, stime",
	    "M"oe"gliche Werte: Nord, NZeit, S"ue"d, SZeit");

    EXPLAIN(ECLIPSE_NAME,
	    "eclipse",
	    "Eklipse");
    EXPLAIN(ECLIPSE_DESC,
	    "Calculates the principal characteristics of an eclipse",
	    "Berechnet die Hauptcharakteristika einer Finsternis");
 /* EXPLAIN(STRING_NAME... */
    EXPLAIN(ECLIPSES_DESC,
	    "Possible values: solar, lunar, all",
	    "M"oe"gliche Werte: Sonne, Mond, Alle");
 /* EXPLAIN(STRING_NAME...*/
    EXPLAIN(ECLVALUE_DESC,
	    "Possible values: type, time, maximum, gamma, u, lunation"
	    " total_semiduration, partial_semiduration, penumbral_semidurationn",
	    "M"oe"gliche Werte: Typ, Zeit, Maximum, Gamma, U, Lunation,"
	    " Totalit"ae"tshalbdauer, Partielle_Halbdauer, Halbschattenhalbdauer");

    EXPLAIN(ECLTYPE_NAME,
	    "eclipsetype",
	    "Eklipsentyp");
    EXPLAIN(ECLTYPE_DESC,
	    "A short description of the eclipse",
	    "Eine kurze Beschreibung der Finsternis");
 /* EXPLAIN(STRING_NAME... */
    EXPLAIN(ECLTYPES_DESC,
	    "The return value from ECLIPSE(...,type)",
	    "Der R"ue"ckgabewert von EKLIPSE(..., Typ)");
    
    EXPLAIN(SEASON_NAME,
	    "season",
	    "Jahreszeit");
    EXPLAIN(SEASON_DESC,
	    "Computes the beginning of the seasons",
	    "Berechnet den Beginn der Jahreszeiten");
 /* EXPLAIN(STRING_NAME... */
    EXPLAIN(SEASONS_DESC,
	    "Possible values: spring, summer, autumn, winter",
	    "M"oe"gliche Werte: Fr"ue"hling, Sommer, Herbst, Winter");
    
    default:
  }

  return("");
}


/*
 * Textuelle Beschreibung des Himmeslskoerpers in Identifizierer umwandeln
 * und die Weichen fuer die Koordinatentransformationen stellen.
 */
static BODY getbody(char *body)
{
  trans.from = HELIO_ECLI_MEAN;
  trans.oldepoch = E_DATE;

  switch (GET2(body)) {		/* die ersten beiden Buchstaben */

    case MKNUM('E','R'):
    case MKNUM('E','A'):
      trans.from = GEO_ECLI_MEAN;
      return trans.body = EARTH;

    case MKNUM('B','A'):		/* planetares Massenzentrum */
      trans.from = BARY_ECLI_MEAN;
      trans.oldepoch = E_J2000;
      return trans.body = BARYCENTER;

     case MKNUM('E','M'):		/* Schwerpunkt Erde-Mond */
      trans.oldepoch = E_J2000;
      return trans.body = EARTHMOON;

    case MKNUM('S','O'):
    case MKNUM('S','U'):
      return trans.body = SUN;

    case MKNUM('M','E'):
      return trans.body = MERCURY;

    case MKNUM('V','E'):
      return trans.body = VENUS;

    case MKNUM('M','A'):
      return trans.body = MARS;

    case MKNUM('J','U'):
      return trans.body = JUPITER;

    case MKNUM('S','A'):
      return trans.body = SATURN;

    case MKNUM('U','R'):
      return trans.body = URANUS;

    case MKNUM('N','E'):
      return trans.body = NEPTUNE;

    case MKNUM('P','L'):
      trans.oldepoch = E_J2000;
      return trans.body = PLUTO;

    case MKNUM('M','O'):
      trans.from = GEO_ECLI_MEAN;
      return trans.body = MOON;

    default:
  }

  return trans.body = -1;
}

#if 0
/*
 * Konvertiert das textuelle Datum in dynamische Zeit TD
 * 
 * Eingabeformate sind:  (# = Zahl, $ = Text)
 *
 * # $ #        Jahr Monatskuerzel Tag(mit Tagesbruchzahl)
 * # $ # #      Jahr Monatskuerzel Tag Stunde(sexagesimal)
 * # $ # #:#:#  Jahr Monatskuerzel Tag Stunde Minute Sekunde
 * # # #        Jahr Monat Tag(mit Tagesbruchzahl)
 * # # # #:#:#  Jahr Monat Tag Stunde Minute Sekunde
 * a            Aktuelle Rechnerzeit (= UT)
 * B#           Besselsche Epoche
 * J#           Julianische Epoche
 * JD #         Julianische Tageszahl
 * MJD #        Modifizierte Julianische Tageszahl
 * T #          Jahrhunderte seit J2000.0 (T = TD = TDT)
 * TDB #        Jahrhunderte seit J2000.0 (baryzentrisch dynamische Zeit)
 */
static double getdate(char *date)
{
  char *ptr = date;
  double t;

  t = strtod(ptr, &ptr);

  if (!*ptr)
    return SC2TD(t);		/* einzelne Zahl: SO-Format */

  if (!*_tscan(&t, date)) {
    if (t >= JDC(-76715589) && t <= JDC(80156580)) {
      return t;
    }
  }
  
  return INVALID;		/* Err:503 */
}
#endif

/*
 * Stellt die Zeitzone ein	(keep it simple!)
 *
 * Rueckgabe ist 1 wenn sich die Zeitzone geaendert hat, sonst 0.
 *
 * Moegliche Werte sind:
 *
 *	UT	(Weltzeit, UTC, GMT)
 *	TD	(dynamische Zeit, Ephemeridenzeit)
 *	System	(Zonenzeit OHNE Sommerzeit)
 *	Aktuell	(Zonenzeit MIT Sommerzeit)
 *
 *	User	(benutzerdefiniert: Zeitdifferenz [h] & Kuerzel)
 */
static int settz(const char *tz)
{
  char *ptr;
  double t, old = tzone;
  static char user[16];

  deltat = tzone = 0;		/* auf Standard (UT) zuruecksetzen */

  switch (GET1(tz)) {		/* der erste Buchstabe entscheidet */

    case 'A':			/* Aktuell */
      tname = atname;
      tzone = atzone;
      break;

    case 'S':			/* System */
      tname = tzname[0];
      tzone = stzone;
      break;

    case 'T':			/* TD */
      tname = "TD";
      deltat = 1E-300;		/* praktisch 0, d.h. nicht benutzen */
      break;

    case 'K':			/* keine */
    case 'N':			/* none */
    case 'U':			/* UT */
      tname = "UT";
      break;

  default:			/* benutzerdefiniert */

    if ((t = strtod(tz, &ptr)) != 0) {	/* User? */
      if (t >= -12 && t <= 12) {		/* korrekt? */
	tname = strncpy(user, ptr, 8);
	tzone = t * (3600 / SEC_PER_CY);	/* Stunden in TD umwandeln */
      }
      else
	tname = INVALID_ARG;		/* mehr als 12h geht nicht! */
    }
    else
      tname = strncpy(user, ptr, 8);  /* nur Namen aendern */
  }

  return (old == tzone)? 0: 1;
}


/*
 * Stellt die Beobachterkoordinaten ein
 *
 * Rueckgabe ist 1 wenn sich der Beobachter geaendert hat, sonst 0.
 *
 * Laengen werden nach Westen positiv gezaehlt. 
 *
 * Die Beobachterhoehe (NN) wird derzeit nicht beruecksichtigt,
 * kann aber leicht eingefuegt werden.
 */
static int setobs(double lon, double lat)
{
  static double ln, lt;
  
  if (ln == lon && lt == lat)	/* veraendert? */
    return 0;

  fun->obs.pos.l = (ln = lon) * radian;
  fun->obs.pos.b = (lt = lat) * radian;

  newobs(&fun->obs);		/* interne Strukturen neu berechnen */

  return 1;
}


/*
 * Auf- und Untergangszeiten vorformatieren
 */
static double rtstime(double t)
{
  if (t < 0 || t > 1)
    return F_INF;		/* bedeutet hier: heute nicht */

  return t;
}


/***** EXPORTIERTE FUNKTIONEN *****/

/*
 * Diese Funktion wird vor allen anderen aufgerufen.
 * 
 * Sie bestimmt die Anzahl der exportierten Funktionen und
 * ist der beste Ort um das Modul zu initialisieren.
 */

SO_CALLBACK( GetFunctionCount, USHORT *nCount)
{
#ifdef __linux__
  onerr = _popup;		/* vorerst noch Rechenfehler melden */
#endif

  *nCount = COUNT_OF_FUNCS;

  setlocale(LC_NUMERIC, "");		/* Dezimalpunkt einstellen */
  
  /* TODO? INI-Datei */

  atname = gettz();			/* Bibliothek initialisieren */

  stzone = timezone * censec;		/* Zeitdifferenz: Standard */
  atzone = tzone;			/* Zeitdifferenz: Aktuell */

  newobs(&fun->obs);
}


#ifdef HAVE_UPDATE
/*
 * Die Funktion liefert immer 0 zurueck. Wird sie zu einem beliebigen Zellwert
 * addiert, dann wird dadurch eine Neuberechnung erzwungen.
 */
SO_CALLBACK( SC_UPDATE, double *r)
{
  *r = 0;
}
#endif
#ifdef INCL_SYSINFO
/*
 * Diese Funktion fehlt leider in StarOffice
 */
SO_CALLBACK( SC_SYSINFO, char *r, char *sub)
{
  int fd, i;
  struct hostent *het;
  struct passwd *pwd;
  struct sysinfo sys;
  struct utsname uts;

  if (*sub == '/') {			/* Datei lesen: /proc/..., /dev/...  */
    if ((fd = open(sub, O_RDONLY)) > 0) {
      i = read(fd, r, 255);
      close(fd);
      if (i > 0) {
	r[i] = '\0';
	return;
      }
    }
  }

  uname(&uts);
  pwd = getpwuid(geteuid());
  het = gethostbyname(uts.nodename);

  if (sysinfo(&sys))
    memset(&sys, 0, sizeof(sys));	/* bei Fehlern `ausnullen' */

  *r = 0;

  switch (GET2(sub)) {

    case MKNUM('B','O'):		/* boot */
      *(short*)r = 0x002f;	/* "/" */	
      return;

    case MKNUM('B','U'):		/* buffered */
      sprintf(r, "%ld", sys.bufferram);
      return;

    case MKNUM('D','O'):		/* domain */
      getdomainname(r, 255);
      return;

    case MKNUM('F','Q'):		/* FQDN */
      if (!het) return;
      SO_StringCopy(r, het->h_name);
      return;

    case MKNUM('H','O'):
      if ((sub[2] & 0xdf) == 'M') {	/* home */
        if (!pwd) return;
	SO_StringCopy(r, pwd->pw_dir);
      }
      
      else				/* host */
	SO_StringCopy(r, uts.nodename);
      return;

    case MKNUM('I','P'):		/* IP */
      if (!het) return;
      if (het->h_length < 4) return;
      sprintf(r, "%d.%d.%d.%d", IP(het,0), IP(het,1), IP(het,2), IP(het,3));
      return;

    case MKNUM('M','A'):		/* machine */
      SO_StringCopy(r, uts.machine);
      return;

    case MKNUM('F','P'):		/* FPU */
      SO_StringCopy(r, fpuoem());
      return;

    case MKNUM('L','O'):		/* load */
      sprintf(r, "%ld", sys.loads[0]);
      return;

    case MKNUM('M','E'):		/* memory */
      sprintf(r, "%ld", sys.freeram);
      return;

    case MKNUM('P','A'):		/* path */
      getcwd(r, 255);
      return;

    case MKNUM('P','R'):		/* procs */
      sprintf(r, "%d", sys.procs);
      return;

    case MKNUM('R','A'):		/* RAM */
      sprintf(r, "%ld", sys.totalram);
      return;

    case MKNUM('R','E'):		/* release */
      SO_StringCopy(r, uts.release);
      return;

    case MKNUM('S','H'):		/* shared */
      sprintf(r, "%ld", sys.sharedram);
      return;

    case MKNUM('S','W'):		/* swaped */
      sprintf(r, "%ld", sys.totalswap - sys.freeswap);
      return;

    case MKNUM('S','Y'):		/* system */
      SO_StringCopy(r, uts.sysname);
      return;

    case MKNUM('U','P'):		/* uptime */
      sprintf(r, "%ld", sys.uptime);
      return;

    case MKNUM('U','S'):		/* user */
      if (!pwd) return;
      SO_StringCopy(r, pwd->pw_name);
      return;

    case MKNUM('V','I'):		/* virtual */
      sprintf(r, "%ld", sys.totalswap);
      return;

    default:
  }

  SO_StringCopy(r, INVALID_ARG);
}
#endif

/**
 ** Ab hier kommen die exportierten Rechenfunktionen
 **/

/*
 * Die Funktion liefert die interne Zahl fuer ein beliebiges Datum.
 *
 * Sie unterliegt nicht den StarOffice Beschraenkungen.
 * (Der Gueltigkeitszeitraum ist von -214748 Mar 1 bis +214749 Feb 28)
 *
 * Eingabeformate sind:  (# = Zahl, $ = Text)
 *
 * # $ #        Jahr Monatskuerzel Tag(mit Tagesbruchzahl)
 * # $ # #      Jahr Monatskuerzel Tag Stunde(sexagesimal)
 * # $ # #:#:#  Jahr Monatskuerzel Tag Stunde Minute Sekunde
 * # # #        Jahr Monat Tag(mit Tagesbruchzahl)
 * # # # #:#:#  Jahr Monat Tag Stunde Minute Sekunde
 * a            Aktuelle Rechnerzeit (= UT)
 * B#           Besselsche Epoche
 * J#           Julianische Epoche
 * JD #         Julianische Tageszahl
 * MJD #        Modifizierte Julianische Tageszahl
 * T #          Jahrhunderte seit J2000.0 (T = TD = TDT)
 * TDB #        Jahrhunderte seit J2000.0 (baryzentrisch dynamische Zeit)
 */
SO_CALLBACK( SC_DATE2, double *r, char *date)
{
  char *ptr = date;

  *r = strtod(ptr, &ptr);

  if (!*ptr)		/* einzelne Zahl: SO-Format */
    return;

  if (!*_tscan(r, date)) {
    if (*r >= JDC(-76715589) && *r <= JDC(80156580)) {
      *r = TD2SC(*r);
      return;
    }
  }
  
  *r = INVALID;		/* Err:503 */
}


/*
 * Die Funktion berechnet den Ostertermin.
 */
SO_CALLBACK( SC_EASTER ,double *r, double *year)
{
  DATE date;

  date.year = *year;

  easter(&date);

  *r = (fabs(*year) < 214749)? jd(&date) - 2415018.5: INVALID;
}


/*
 * Winkel: Grad, Minute, Sekunde(nbruchteile) formatiert ausgeben
 */
SO_QUIRKSCALL( SC_DMS, char *r, double *a, double *n, char *fmt)
{
  char *units = (strlen(fmt) < 3)? DG "\'\"": fmt;
  
  SO_StringCopy(r, aprint(*a, units, (int)*n, 0));
}


/*
 * Stundenwinkel: Stunde, Minute, Sekunde(nbruchteile) formatiert ausgeben
 */
SO_QUIRKSCALL( SC_HMS, char *r, double *a, double *n, char *fmt)
{
  char *units = (strlen(fmt) < 3)? "hms": fmt;
  
  SO_StringCopy(r, aprint(dgr2(*a) * (1/15.0), units, (int)*n, 0));
}


/*
 * Datum und/oder Zeit formatiert ausgeben
 *
 * Format-String:
 *
 *	[T][J][M|M$][T#|T][J][H][M][S#]
 *
 * Dieselben Inkonsistenzen liegen auch in StarCalc vor ;-))
 *
 * TT.MM.JJJJ aber HH.MM.SS (nicht SS.MM.SS; ist MM nun Monat oder Minute?)
 *
 */
SO_CALLBACK( SC_DATIME, char *r, double *t, char *fmt)
{
  char c, *ptr = fmt;
  DFORMAT dfm = P_UNIT;		/* externe Einheiten: hier Zeittrenner */

  dprint_unit = ":: ";		/* Standard StarCalc-Zeittrenner */ 

  if (*t < -79130607.5 || *t > 77741561.5) {
    SO_StringCopy(r, INVALID_ARG);			/* Err:502 */
    return;
  }

  for (;;) {

  /* Die Abfrage-Reihenfolge ist hier entscheidend */

    if (GET1(ptr) == 'D' || GET1(ptr) == 'T') {
      dfm |= P_DAY1;
      ptr++;
    }

    if (GET1(ptr) == 'Y' || GET1(ptr) == 'J') {
      dfm |= P_YEAR1;
      ptr++;
    }

    if (GET1(ptr) == 'M') {
      if (*(++ptr) == '$') {
	dfm |= P_MNAME;
	ptr++;
      }
      else
	dfm |= P_MONTH;
    }

    if (GET1(ptr) == 'D' || GET1(ptr) == 'T') {
      c = *(++ptr);
      if (c >= '0' && c <= '7') {		/* Tagesbruchzahl? */
	dfm |= P_DFRAC | (c & 0x07);
	break;				/* dann keine HMS... mehr */
      }
      else
	dfm |= P_DAY2;
    }
  
    if (GET1(ptr) == 'Y' || GET1(ptr) == 'J') {
      dfm |= P_YEAR2;
      ptr++;
    }

    if (GET1(ptr) == 'H') {
      dfm |= P_HOUR;
      ptr++;
    }

    if (GET1(ptr) == 'M') {
      dfm |= P_MIN;
      ptr++;
    }

    if (GET1(ptr) == 'S') {
      dfm |= P_SEC;
      c = *(++ptr);
      if (c >= '0' && c <= '7') {		/* Sekundenbruchteile? */
	dfm |= c & 0x07;
	dprint_unit = "::,";
      }
    }

    break;
  }
  
  SO_StringCopy(r, dprint(SC2TD(*t), dfm));
}


/*
 * Stellt die Zeitzone ein (fragt sie ab)
 */
SO_CALLBACK( SC_TIMEZONE, char *r, char *tz)
{
  if (*tz)
    settz(tz);
  
  SO_StringCopy(r, tname);
}


/*
 * Berechnung der Koordinaten des Himmelskoerpers
 */
SO_QUIRKSCALL( SC_POSITION, double *r, double *date, char *body, char *sub,
			  char *cntr, char *plan, char *type, char *eqnx,
			  double *lon, double *lat)
{
  double e, t = SC2TD(*date);
  BODY b = trans.body;
  COORD c = 0;
  static double g, T = UINIT;
  static CARTESIC rec;
  static SPHERIC pos;

  if (getbody(body) == -1) {		/* nur zugelassene Himmelskoerper */
    *r = INVALID;
    return;
  }

  switch (GET1(cntr)) {		/* Ursprung bestimmen */
    
    case 'B':			/* baryzentrisch */
      c |= C_BARY;
      break;

    case 'H':			/* heliozentrisch */
      c |= C_HELIO;
      break;

    case 'G':			/* geozentrisch */
      c |= C_GEO;
      break;

    case 'T':			/* topozentrisch */
      c |= C_TOPO;
      break;

    default:
      *r = INVALID;
      return;
  }

  switch (GET2(plan)) {		/* Bezugsebene bestimmen */
    
    case MKNUM('E','C'):	/* ekliptikal */
    case MKNUM('E','K'):
      c |= C_ECLI;
      break;

    case MKNUM('A' ,'E'):	/* aequatorial */
    case MKNUM('E','Q'):
    case MKNUM(_AE,'Q'):
      c |= C_EQUA;
      break;

    case MKNUM('H','O'):	/* horizontal */
      c |= C_HORI;
      break;

    default:
      *r = INVALID;
      return;
  }

  switch (GET1(type)) {		/* Koordinatentyp bestimmen */
    
    case 'M':			/* mittel */
      c |= C_MEAN;
      break;

    case 'T':			/* wahr */
    case 'W':
      c |= C_TRUE;
      break;

    case 'A':			/* scheinbar */
    case 'S':
      c |= C_APPA;
      break;

    default:
      *r = INVALID;
      return;
  }

  if (GET1(eqnx) == 'D')	/* Aequinoktium der Koordinaten bestimmen */
    e = t;
  else if (*_tscan(&e, eqnx)){
    *r = INVALID;
    return;			/* ungueltiges Argument */
  }

		/* schon berechnet? */

  if (setobs(*lon, *lat) || b != trans.body ||
      c != trans.to || e != trans.neu || T != t) {
    T = t;

    trans.neu = e;		/* Koordinatentransformation vorbereiten */
    trans.to = c;

    if (fabs(e - t) < 1e-10)
      trans.newepoch = E_DATE;		/* Datum */
    else if (e == 0)
      trans.newepoch = E_J2000;		/* J2000 */
    else if (e == BESSEL(1950))
      trans.newepoch = E_B1950;		/* B1950 */
    else
      trans.newepoch = E_OTHER;		/* alle anderen */

    newdate(t, fun);		/* Grunddaten berechnen */

    switch (trans.body) {	/* Bahnposition bestimmen */
      case BARYCENTER:
      case SUN: 
      case EARTH:
        rec.x = rec.y = rec.z = 0;
        break;

      case EARTHMOON:
      case MOON:
	position(trans.body, t, &rec);
	g = betrag(&rec);			/* geometrische Entfernung */
	if (!ISMEAN(trans.to))
	  position(trans.body, t - g * CONST_LTCOR, &rec);  /* Lichtlaufzeit */
	break;

      default:
        if (ISHELIO(trans.to)) {
          position(trans.body, t, &rec);	/* geometrische Position */
          g = betrag(&rec);
        }		     /* Bewegung waehrend Lichtlaufzeit retardieren */
        else
          g = astrometric(&trans, t, &rec);	/* astrometrische Position */
    }

    spheric(transform(&trans, &rec, &rec), &pos);	/* transformieren */

    if (trans.body < MERCURY)
      g = pos.r;		/* geometrische = astrometrische Entfernung */
  }

  switch (GET1(sub)) {	  /* Daten ausgeben */

    case 'G':			/* wahre (geometrische) Entfernung */
    case 'W':
    case 'T':
      *r = g;
      break;

    case 'R':			/* sphaerische Koordinate */
      *r = pos.r;
      break;
      
    case 'L':
      *r = rd2dg2(pos.l);
      break;
      
    case 'B':
      *r = rd2dg(pos.b);
      break;
      
    case 'X':			/* rechtwinklige Koordinate */
      *r = rec.x;
      break;
      
    case 'Y':
      *r = rec.y;
      break;

    case 'Z':
      *r = rec.z;
      break;

    default:
      *r = INVALID;
  }
}


/*
 * Berechnung der Zeiten der Auf- und Untergaenge
 */
SO_QUIRKSCALL( SC_VISIBLE, double *r, double *date, char *body, char *sub,
			 char *tz, double *lon, double *lat)
{
  int i = settz(tz);
  double t = SC2TD(*date);
  BODY b = trans.body;
  CARTESIC rec;
  SPHERIC pos, sph[3];
  static double T = UINIT;
  static RISET rts, civil, naut, astro;

  if (getbody(body) == -1) {		/* nur zugelassene Himmelskoerper */
    *r = INVALID;
    return;
  }
	/* schon berechnet? */

  if (setobs(*lon, *lat) || i || b != trans.body || T != t ) {
    T = t;

    trans.to = GEO_EQUA_APPA;	/* scheinbare Koordinaten benutzen */
    trans.newepoch = E_DATE;

    precision = MEDIUM;		/* volle Genauigkeit ist hier ueberfluessig */
    
    for (i = 0, t = daybeg(T - censec); i < 3; i++, t += cenday) {

      newdate(t, fun);		/* Grunddaten berechnen */

      memmove(&sph[0], &sph[1], 2*sizeof(SPHERIC));	/* Daten versetzen */

      if (trans.body == SUN)		/* Sonnenposition ist schon bekannt */
	rec.x = rec.y = rec.z = 0;
      else
	spheric(position(trans.body, t, &rec), &pos);	/* Bahn-Ort */

      spheric(transform(&trans, &rec, &rec), &sph[2]);	/* geo-ecli */

      sph[2].r = fun->tim.last;		/* Ortssternzeit eintragen */
    }

    switch (trans.body) {		/* Standardhoehen bestimmen */
      case SUN:
        rts.h0 = -0.0046542;		/* Radius */

	civil.h0 = -0.1047198;		/* -6 Grad, buergerliche Daemmerung */
	riset(sph, &civil);

	naut.h0 = -0.2094395;		/* -12 Grad, nautische Daemmerung */
	riset(sph, &naut);

	astro.h0 = -0.3141593;		/* -18 Grad, astronomische Daemmerung */
	riset(sph, &astro);
        break;

      case MOON:		
        rts.h0 = 3.10171e-05 / pos.r;	/* Radius und Parallaxe */
        break;

      default:
        rts.h0 = 0;
    }
					    /* Refraktion und Kimmtiefe */
    rts.h0 += refrac(0); //+ dip(fun->obs.pos.r);

    riset(sph, &rts);			/* Zeiten berechnen */

    precision = HIGH;
  }

  switch (GET2(sub)) {

    case MKNUM('A','U'):	/* Aufgang */
    case MKNUM('R','I'):
      *r = rtstime(rts.rise);
      break;

    case MKNUM('D','U'):	/* Durchgang */
    case MKNUM('T','R'):
      *r = rtstime(rts.trans);
      break;

    case MKNUM('U','N'):	/* Untergang */
    case MKNUM('S','E'):
      *r = rtstime(rts.set);
      break;

    case MKNUM('K','U'):	/* Kulmination */
    case MKNUM('C','U'):
      *r = rts.alti * degree;
      break;

    default:
      *r = INVALID;
  }

  if (trans.body == SUN) {	/* Daemmerungen */

    switch (GET2(sub)) {

      case MKNUM('B','B'):	/* buergerlich, Beginn */ 
      case MKNUM('C','B'): 
	*r = rtstime(civil.rise);
	break;

      case MKNUM('B','E'):	/* buergerlich, Ende */ 
      case MKNUM('C','E'): 
	*r = rtstime(civil.set);
	break;

      case MKNUM('N','B'):	/* nautisch, Beginn */ 
        *r = rtstime(naut.rise);
	break;

      case MKNUM('N','E'):	/* nautisch, Ende */ 
	*r = rtstime(naut.set);
	break;

      case MKNUM('A','B'):	/* astronomisch, Beginn */ 
	*r = rtstime(astro.rise);
	break;

      case MKNUM('A','E'):	/* astronomisch, Ende */ 
	*r = rtstime(astro.set);
	break;
    }
  }
}


/*
 * Berechnung der physischen Epemeriden
 */
SO_QUIRKSCALL( SC_PHYSICAL, double *r, double *date, char *body, char *sub)
{
  double t = SC2TD(*date);
  static double T = UINIT;
  static PHYSICAL phys;
  
  if (getbody(body) == -1) {		/* nur zugelassene Himmelskoerper */
    *r = INVALID;
    return;
  }

  if (phys.body != trans.body || T != t) {	/* schon einmal berechnet? */
    T = t;

    trans.to = GEO_EQUA_APPA;	/* scheinbare Koordinaten benutzen */
    trans.newepoch = E_DATE;

    precision = MEDIUM;		/* volle Genauigkeit ist hier ueberfluessig */

    newdate(t, fun);		/* wird immer benoetigt */

    switch (phys.body = trans.body) {

      case SUN:		/* hier mehr oder weniger sinnvolle Werte eintragen */
	magnitude(&phys);

	phys.i = phys.q = phys.elon = phys.psan = 0;
	phys.k = 1;
	
	phys.dgeo = fun->sun.pos.r;
	phys.para = asin(shape[EARTH].radi / fun->sun.pos.r);
	phys.requ = phys.rpol = asin(shape[SUN].radi / fun->sun.pos.r);
        break;

      case MOON:				/* Mond */
        geo_hel_ecl(moonpos(t, &phys.hel), &phys.hel);

        physical(&phys);
	phys.mag = -12.55;		/* TODO */	
        break;

      case MERCURY ... PLUTO:			/* Merkur - Pluto */
        position(trans.body, t, &phys.hel);

        physical(&phys);
        break;

      default:
        *r = INVALID;
	return;
    }

    precision = HIGH;
  }

  switch (GET2(sub)) {

    /* Alle Winkel werden in Grad ausgegeben */ 

    case MKNUM('B','E'):		/* Defekt, Beleuchtungsdefekt */
    case MKNUM('D','E'):
    case MKNUM('Q','\0'):
      *r = phys.q * degree;
      break;

    case MKNUM('E','L'):		/* Elongation */
      *r = phys.elon * degree;
      *r = (phys.psan < 0)? -*r: *r;	/* oestlich: < 0; westlich: > 0 */
      break;

    case MKNUM('H','E'):		/* Helligkeit, Magnitude */
    case MKNUM('M','A'):
      *r = phys.mag;
      break;

    case MKNUM('L','I'):		/* Lichtzeit */
    case MKNUM('L','T'):
    case MKNUM('L','Z'):
      *r = phys.dgeo * CONST_LTIME;	/* in Sekunden */
      break;

    case MKNUM('P','H'):		/* Phase */
    case MKNUM('K','\0'):
      *r = phys.k;
      break;

    case MKNUM('I','\0'):		/* Phasewinkel i */
      *r = phys.i * degree;
      break;

    case MKNUM('P','A'):		/* Parallaxe */
      *r = phys.para * degree;
      break;

    case MKNUM('P','O'):		/* Positionswinkel der Sonne */
      *r = rd2dg2(phys.psan);
      break;

    case MKNUM('P','R'):		/* Polradius (Grad) */
      *r = phys.rpol * degree;
      break;

    case MKNUM('R','A'):		/* Radius, Aequatorradius (Grad)  */
      *r = phys.requ * degree;
      break;

    default:
      *r = INVALID;
  }
}


/*
 * Berechnung der kartografischen Koordinaten fuer den Himmelskoerper
 */
SO_QUIRKSCALL( SC_ROTATION, double *r, double *date, char *body, char *coor,
			  char *sub, double *lon, double *lat)
{
  double d, t = SC2TD(*date);
  BODY b = trans.body;
  CARTESIC hel, geo, pos;
  PLANGRAF *plan;
  ROTATION rot;
  static double T = UINIT;
  static PLANGRAF pla, sol, top;

  if (getbody(body) == -1) {		/* nur zugelassene Himmelskoerper */
    *r = INVALID;
    return;
  }

  if (setobs(*lon, *lat) || b != trans.body || T != t) { /* schon berechnet? */
    T = t;

    trans.to = GEO_EQUA_APPA;	/* scheinbare Koordinaten benutzen */
    trans.newepoch = E_DATE;

    precision = MEDIUM;		/* volle Genauigkeit ist hier ueberfluessig */

    newdate(t, fun);		/* wird immer benoetigt */

    switch (trans.body) {
      case SUN:					/* Sonne */
        rot_sun(t, &rot);
        abersun(EARTH, &fun->sun.rec);		/* Aberration des Lichts */
        rotation(&fun->sun.rec, &rot, &pla);

	memset(&sol, 0, sizeof(PLANGRAF));	/* sinnvolle Werte eintragen */
	memcpy(&top, &pla, sizeof(PLANGRAF));
        break;

      case MOON:				/* Mond */
        moonpos(t, &geo);
        rot_moon(t, &rot);

        rotation(&geo, &rot, &pla);	/* geozentrische Libration */

        geo_top_ecl(&geo, &pos);
        rotation(&pos, &rot, &top);	/* topozentrische Libration */

        geo_hel_ecl(&geo, &hel);
        abersun(MOON, &pos);		/* Aberration der Sonne vom Mond aus */
        rotation(&hel, &rot, &sol);	/* heliozentrische Libration */
        break;

      case MERCURY ... PLUTO:			/* Merkur - Pluto */
	d = CONST_LTCOR * astrometric(&trans, t, &hel);

        orientation(trans.body, t - d, &rot);	/* Lage der Rotationsachse */

        hel_geo_ecl(&hel, &geo);		/* geozentr. Koordinaten */
        rotation(&geo, &rot, &pla);		/* planetogr. Erdkoordinaten */

        abersun(trans.body, &hel);		/* Aberration der Sonne */
        rotation(&hel, &rot, &sol);		/* planetogr. Sonnenkoord. */

	/* Phasenkorrektur der grossen Planeten hier ignorieren */

	memcpy(&top, &pla, sizeof(PLANGRAF));
        break;

      default:
        *r = INVALID;
	return;
    }

    precision = HIGH;
  }

  switch (GET1(coor)) {		/* Koordinatensystem auswaehlen */

    case 'G':			/* geozentrische Orientierung */
      plan = &pla;
      break;

    case 'H':			/* heliozentrische Orientierung */
      plan = &sol;
      break;

    case 'T':			/* topozentrische Orientierung */
      plan = &top;
      break;

    default:
      *r = INVALID;
      return;
  }

  switch (GET2(sub)) {		/* Koordinate abrufen */

    /* Die Winkel werden entsprechend astronomischer Gewohnheit formatiert */

    case MKNUM('P','O'):	/* Positionswinkel der Achse */
      *r = (trans.body == SUN)?
	rd2dg(plan->pa): rd2dg2(plan->pa);
      return;

    case MKNUM('L','\0'):	/* planetozentrische Laenge */
      *r = (trans.body == MOON)?
	rd2dg(plan->l): rd2dg2(plan->l);
      return;

    case MKNUM('L','\x11'):	/* planetografische Laenge (System I) */
      *r = rd2dg2(plan->lam);
      return;		/* Zentralmeridian im Aequatorbereich */

    case MKNUM('L','\x12'):	/* planetografische Laenge (System II) */
      *r = rd2dg2(plan->lam2);
      return;		/* Zentralmeridian der detailreichen Baender */

    case MKNUM('L','\x13'):	/* planetografische Laenge (System III) */
      *r = rd2dg2(plan->lam3);
      return;		/* Zentralmeridian der Radiostrahlung */

    case MKNUM('B','\0'):	/* planetozentrische Breite */
      *r = rd2dg(plan->b);
      return;

    case MKNUM('B','E'):	/* planetografische Breite */
      *r = rd2dg(plan->bet);
      return;

      /* fuer den Mond zusaetzlich Colongitude und Terminator */

    case MKNUM('C','O'):	/* Colongitude */
      *r = rd2dg(_pi_2 - sol.l);
      return;

    case MKNUM('L','I'):	/* Terminator, Lichtgrenze */
    case MKNUM('T','E'):
      *r = -rd2dg(asin(fsign(sin(sol.l)) * cos(sol.l)));
      return;

    default:
      *r = INVALID;
      return;
  }
}


/*
 * Berechnung der Besselschen Tageszahlen, Ekliptikschiefe und Nutation
 */
SO_CALLBACK( SC_BESSELDN, double *r, double *date, char *sub)
{
  double t = SC2TD(*date);
  static double de, dp, o, T = UINIT;
  static BESDAYNUM bdn;

  if (T != t) {
    newdate(T = t, fun);

    de = fun->nut.de;
    dp = fun->nut.dp;
    o = fun->obl.te;

    bdaynum(&bdn);
  }

  switch (GET1(sub)) {

   case 'A':
     *r = bdn.a * degree;
     return;

   case 'B':
     *r = bdn.b * degree;
     return;

   case 'C':
     *r = bdn.c * degree;
     return;

   case 'D':
     *r = bdn.d * degree;
     return;

   case 'E':
     *r = bdn.e * degree;
     return;

   case 'M':
     *r = bdn.mn;
     return;

   case 'T':
     *r = bdn.te;
     return;

   case 'O':
   case 'S':
     *r = o * degree;
     return;

   case 'X':
     *r = dp * degree;
     return;

   case 'Y':
     *r = de * degree;
     return;

   default:
     *r = INVALID;
     return;
  }
}


/*
 * Berechnung der Lunation, d.h. der Anzahl der Neumonde seit J2000
 */
SO_CALLBACK( SC_LUNATION, double *r, double *date)
{
  *r = floor(SC2TD(*date) * 1236.853088);  /* Mittlere Lunation ermitteln */ 
}


/*
 * Berechnet zu einer Lunation die entsprechende Zeitpunkte der Mondphasen 
 */
SO_QUIRKSCALL( SC_MOONPHASE, double *r, double *lun, char *sub, char *tz)
{
  double d;
  static PHASE phase;
  static REVOLUTION l = 0x7fffffff;

  if (*lun != l)		/* wurden die Phasen schon berechnet? */ 
     moonphase(l = *lun, &phase);

  settz(tz);			/* Zeitzone einstellen */

  d = dt(phase.neu) + tzone;

  switch (GET2(sub)) {		/* die ersten beiden Buchstaben */

    case MKNUM('N','E'):	/* Neumond, newmoon */
      *r = TD2SC(phase.neu - d);
      break;

    case MKNUM('E','R'):	/* Erstesviertel, firstquater */
    case MKNUM('F','I'):
      *r = TD2SC(phase.first - d);
      break;

    case MKNUM('V','O'):	/* Vollmond, fullmoon */
    case MKNUM('F','U'):
      *r = TD2SC(phase.full - d);
      break;

    case MKNUM('L','E'):	/* Letztesviertel, lastquater */
    case MKNUM('L','A'):
      *r = TD2SC(phase.last - d);
      break;

    default:
      *r = INVALID;
  }
}


/*
 * Berechnung der Apsiden, d.h. der Anzahl der Konjunktionen des Mondes
 * mit seinem Perigaeum seit J2000
 */
SO_CALLBACK( SC_APSIDE, double *r, double *date)
{
  *r = floor(SC2TD(*date) * 1325.552418); /* Mittlere Anomale ermitteln */ 
}


/*
 * Berechnet zu einer Apside die entsprechenden Zeitpunkte und Entfernungen 
 */
SO_QUIRKSCALL( SC_MOONAPSIDE, double *r, double *apsi, char *sub, char *tz)
{
  double d;
  static APSIDE apside;
  static REVOLUTION a = 0x7fffffff;

  if (*apsi != a)		/* schon berechnet? */ 
     moonapside(a = *apsi, &apside);

  settz(tz);			/* Zeitzone einstellen */

  d = dt(apside.ptime) + tzone;

   switch (GET2(sub)) {		/* die ersten beiden Buchstaben */

    case MKNUM('P','E'):	/* Perigaeum, perigee */
      *r = (RADI_EARTH * UNIT_DIST) / sin(apside.peri);		/* in m */
      break;

    case MKNUM('P','Z'):	/* PZeit, ptime */
    case MKNUM('P','T'):
      *r = TD2SC(apside.ptime - d);
      break;

    case MKNUM('A','P'):	/* Apogaeum, apogee */
      *r = (RADI_EARTH * UNIT_DIST) / sin(apside.apo);		/* in m */
      break;

    case MKNUM('A','Z'):	/* AZeit, atime */
    case MKNUM('A','T'):
      *r = TD2SC(apside.atime - d);
      break;

    default:
      *r = INVALID;
  }
}


/*
 * Berechnung des Knoten, d.h. der Anzahl der Konjunktionen des Mondes
 * mit dem aufsteigenden Knoten seit J2000
 */
SO_CALLBACK( SC_NODE, double *r, double *date)
{
  *r = floor(SC2TD(*date) * 1342.227826);
}


/*
 * Berechnet zu einem Knoten die entsprechenden Zeitpunkte
 */
SO_QUIRKSCALL( SC_MOONNODE, double *r, double *n, char *sub, char *tz)
{
  double d;
  static NODE node;
  static REVOLUTION cend = 0x7fffffff;

  if (*n != cend)		/* schon berechnet? */ 
     moonnode(cend = *n, &node);

  settz(tz);			/* Zeitzone einstellen */

  d = dt(node.asc) + tzone;

  switch (GET2(sub)) {		/* die ersten beiden Buchstaben */

    case MKNUM('A','U'):	/* aufsteigend, ascending */
    case MKNUM('A','S'):
      *r = TD2SC(node.asc - d);
      break;

    case MKNUM('A','B'):	/* absteigend, descending */
    case MKNUM('D','E'):
      *r = TD2SC(node.des - d);
      break;

    default:
      *r =  INVALID;
  }
}


/*
 * Berechnung der Trope, d.h. der Anzahl der Konjunktionen des Mondes
 * mit dem Fruehlingspunkt (Tagundnachtgleiche, Aequinoktium) seit J2000
 */
SO_CALLBACK( SC_TROPIC, double *r, double *date)
{
  *r = floor(SC2TD(*date) * 1336.8555226);  /* Mittlere Trope ermitteln */ 
}


/*
 * Berechnet zu einer Trope die entsprechende Zeitpunkte und Deklinationen 
 */
SO_QUIRKSCALL( SC_MOONMAXDECL, double *r, double *equ, char *sub, char *tz)
{
  double d;
  static DECLMAX declmax;
  static REVOLUTION decl = 0x7fffffff;

  if (*equ != decl)		/* schon berechnet? */ 
     moondeclmax(decl = *equ, &declmax);

  settz(tz);			/* Zeitzone einstellen */

  d = dt(declmax.ntime) + tzone;

  switch (GET2(sub)) {		/* die ersten beiden Buchstaben */

    case MKNUM('N','O'):	/* Nord, north */
      *r = declmax.ndecl * degree;
      break;

    case MKNUM('N','Z'):	/* NZeit, ntime */
    case MKNUM('N','T'):
      *r = TD2SC(declmax.ntime - d);
      break;

    case MKNUM('S','U'):	/* Sued, south */
    case MKNUM('S',_UE):
    case MKNUM('S',_ue):
    case MKNUM('S','O'):
      *r = declmax.sdecl * degree;
      break;

    case MKNUM('S','Z'):	/* SZeit, stime */
    case MKNUM('S','T'):
      *r = TD2SC(declmax.stime - d);
      break;

     default:
      *r = INVALID;
  }
}


/*
 * Berechnet zu einer Lunation die entsprechende Finsternis
 *
 * Wenn lun keine ganze Zahl ist, wird mit der Suche bei den
 * Mondfinsternissen begonnen (unterbrochene Suche fortfuehren).
 *
 * ecli.e: 0 = Sonnenfinsternis, 1 = Mondfinsternis suchen
 */
SO_QUIRKSCALL( SC_ECLIPSE, double *r, double *lun, char *typ,
			 char *sub, char *tz)
{
  int x, e;
  static double lu = UINIT;
  static ECLIPSE ecli;
  static REVOLUTION l;

  switch (GET1(typ)) {		/* der erste Buchstabe ist wichtig */

    case 'S':		/* Sonne, sun, solar */
      e = 0;
      x = 0;
      break;

    case 'L':		/* Mond, moon, lunar */
    case 'M':
      e = 1;
      x = 0;
      break;

    case 'A':		/* Alle, both (bedeutet also die naechste) */
    case 'B':
    default:
      e = frac(*lun)? 0: 1;	/* bei Sonne oder Mond beginnen? */
      x = 1;
  }

  settz(tz);			/* Zeitzone einstellen */

  if (lu != *lun || e != ecli.e) {		/* schon berechnet? */
    l = ifloor(lu = *lun);	/* Zahl wieder glaetten */
    ecli.e = e;

				/* die naechste Finsternis suchen */
    do {
      ecli.e ^= x;		/* tricky: Bit 0 nach Bedarf umpolen */

      eclipse(l, &ecli);

      l += ecli.e | (x ^ 1);	/* nur am Ende einer Lunation weiterzaehlen */

    } while (!ecli.type);
  }

  switch (GET2(sub)) {		/* die ersten beiden Buchstaben */

    case MKNUM('T','Y'):	/* Typ */
      *r = (double)(ecli.e? -1.0 * ecli.type: ecli.type);
      break;

    case MKNUM('Z','E'):	/* Zeitpunkt der Finsternis */
    case MKNUM('T','I'):
      *r = TD2SC(ecli.tmax - dt(ecli.tmax) - tzone);
      break;

    case MKNUM('G','R'):	/* Groesse, Maximum der Finsternis */
    case MKNUM('M','A'):
      *r = ecli.max < 0? ecli.penmax: ecli.max;
      break;

    case MKNUM('T','O'):	/* Totale Phase */
      *r = ecli.type >= TOTAL? ecli.total * daycen: 0;
      break;

    case MKNUM('P','A'):	/* Partielle Phase */
      *r = ecli.type >= PARTIAL? ecli.partial * daycen: 0;
      break;

    case MKNUM('H','A'):	/* Halbschattenphase */
    case MKNUM('P','E'):
      *r = (ecli.e == 1)? ecli.penumbral * daycen: 0;
      break;

    case MKNUM('G','A'):	/* Gamma, g */
    case MKNUM('G','\0'):
      *r = ecli.g;
      break;

    case MKNUM('U','\0'):	/* u */
      *r = ecli.u;
      break;

    case MKNUM('L','U'):	/* Lunation */
      *r = (double)l + ((x && ecli.e == 0)? 0.5: 0);
      break;

    default:
      *r = INVALID;
  }
}


/*
 * Textuelle Beschreibung einer Finsternis abrufen
 */
SO_CALLBACK( SC_ECLTYPE, char *r, double *type)
{
  int t = (int)fabs(*type);
  const char **ptr;
  
  static const char *eng[] = {
    " (nc)",
    "Penumbral",
    "Partial",
    "Total",
    "Annular",
    "Annular-total",
    " solar eclipse",
    " lunar eclipse",
  };

  static const char *ger[] = {
    " (nz)",
    "Halbschatten",
    "Partielle",
    "Totale",
    "Ringf"oe"rmige",
    "Ringf"oe"rmig-totale",
    " Sonnenfinsternis",
    " Mondfinsternis",
  };

  switch (_nLanguage) {

    case LANGUAGE_GERMAN:
      ptr = ger;
      break;

    default:
      ptr = eng;
  }

  SO_StringCopy(r, ptr[t & 0x07]);
  strcat(r, ptr[(*type < 0)? 7: 6]);

  if (t > 0x07)
    strcat(r, ptr[0]);
}


/*
 * Jahreszeiten
 */
SO_QUIRKSCALL( SC_SEASON, double *r, double *year, char *sub, char *tz)
{
  double d;
  static double y;
  static SEASON seas;

  if (*year != y)		/* schon berechnet? */
    season((y = *year) - 2000, &seas);

  settz(tz);			/* Zeitzone einstellen */

  d = dt(seas.summer) + tzone;

  switch (GET2(sub)) {		/* die ersten beiden Buchstaben */

    case MKNUM('F','R'):	/* Fruehling, spring */
    case MKNUM('S','P'):
      *r = TD2SC(seas.spring - d);
      break;
      
    case MKNUM('S','O'):	/* Sommer, summer */
    case MKNUM('S','U'):
      *r = TD2SC(seas.summer - d);
      break;

    case MKNUM('H','E'):	/* Herbst, autumn */
    case MKNUM('A','U'):
      *r = TD2SC(seas.autumn - d);
      break;

    case MKNUM('W','I'):	/* Winter */
      *r = TD2SC(seas.winter - d);
      break;

     default:
       *r = INVALID;
  }
}
/*
 * Hier enden die Rechenfunktionen
 */


/*
 * Diese Funktion gibt Auskunft ueber die Zahl der Argumente, Datentypen und
 * Namen der exportierten Funktionen
 *
 * Folgende Parmeter muessen versorgt werden
 *
 * USHORT *nNo:	        Nummer der Funktion zwischen 0 und nCount - 1  
 * char   *pFuncName:   Name der Funktion, die in der SharedLibrary (DLL)
 *			verwendet werden soll
 * USHORT *nParamCount: Anzahl aller Parameter (maximal 16) einschliesslich
 *			des Rueckgabewertes
 * ParamType *peType:   Zeiger auf ein Array mit den Datentypen aller Parameter
 *			Es muessen alle 16 moeglichen Parameter enthalten sein!
 * char *pInternalName: Name der Funktion, die der StarCalc-User sehen soll  
 */
SO_QUIRKSCALL( GetFunctionData, USHORT *nNo, char *pFuncName, USHORT *nParamCount,
                               ParamType *peType, char *pInternalName)
{
  switch(*nNo) {
#ifdef HAVE_UPDATE
    case FUNC_UPDATE:
      SO_StringCopy(pInternalName, getText(UPDATE_NAME));
      SO_StringCopy(pFuncName,     Q(SC_UPDATE));
      peType[0] = PTR_DOUBLE;
      *nParamCount=1;
      break;
#endif
#ifdef INCL_SYSINFO
    case FUNC_SYSINFO:
      SO_StringCopy(pInternalName, getText(SYSINFO_NAME));
      SO_StringCopy(pFuncName,     Q(SC_SYSINFO));
      peType[0] = PTR_STRING;
      peType[1] = PTR_STRING;
      *nParamCount=2;
      break;
#endif

    case FUNC_DATE2:
      SO_StringCopy(pInternalName, getText(DATE2_NAME));
      SO_StringCopy(pFuncName,     Q(SC_DATE2));
      peType[0] = PTR_DOUBLE;
      peType[1] = PTR_STRING;
      *nParamCount=2;
      break;

    case FUNC_EASTER:
      SO_StringCopy(pInternalName, getText(EASTER_NAME));
      SO_StringCopy(pFuncName,     Q(SC_EASTER));
      peType[0] = PTR_DOUBLE;
      peType[1] = PTR_DOUBLE;
      *nParamCount=2;
      break;

    case FUNC_DMS:
      SO_StringCopy(pInternalName, getText(DMS_NAME));
      SO_StringCopy(pFuncName,     Q(SC_DMS));
      peType[0] = PTR_STRING;
      peType[1] = PTR_DOUBLE;
      peType[2] = PTR_DOUBLE;
      peType[3] = PTR_STRING;
      *nParamCount=4;
      break;

    case FUNC_HMS:
      SO_StringCopy(pInternalName, getText(HMS_NAME));
      SO_StringCopy(pFuncName,     Q(SC_HMS));
      peType[0] = PTR_STRING;
      peType[1] = PTR_DOUBLE;
      peType[2] = PTR_DOUBLE;
      peType[3] = PTR_STRING;
      *nParamCount=4;
      break;

    case FUNC_DATIME:
      SO_StringCopy(pInternalName, getText(DATIME_NAME));
      SO_StringCopy(pFuncName,     Q(SC_DATIME));
      peType[0] = PTR_STRING;
      peType[1] = PTR_DOUBLE;
      peType[2] = PTR_STRING;
      *nParamCount=3;
      break;

    case FUNC_TIMEZONE:
      SO_StringCopy(pInternalName, getText(TIMEZONE_NAME));
      SO_StringCopy(pFuncName,     Q(SC_TIMEZONE));
      peType[0] = PTR_STRING;
      peType[1] = PTR_STRING;
      *nParamCount=2;
      break;
	 
    case FUNC_POSITION:
      SO_StringCopy(pInternalName, getText(POSITION_NAME));
      SO_StringCopy(pFuncName,     Q(SC_POSITION));
      peType[0] = PTR_DOUBLE;
      peType[1] = PTR_DOUBLE;
      peType[2] = PTR_STRING;
      peType[3] = PTR_STRING;
      peType[4] = PTR_STRING;
      peType[5] = PTR_STRING;
      peType[6] = PTR_STRING;
      peType[7] = PTR_STRING;
      peType[8] = PTR_DOUBLE;
      peType[9] = PTR_DOUBLE;
      *nParamCount=10;
      break;
	 
    case FUNC_VISIBLE:
      SO_StringCopy(pInternalName, getText(VISIBLE_NAME));
      SO_StringCopy(pFuncName,     Q(SC_VISIBLE));
      peType[0] = PTR_DOUBLE;
      peType[1] = PTR_DOUBLE;
      peType[2] = PTR_STRING;
      peType[3] = PTR_STRING;
      peType[4] = PTR_STRING;
      peType[5] = PTR_DOUBLE;
      peType[6] = PTR_DOUBLE;
      *nParamCount=7;
      break;
	 
    case FUNC_PHYSICAL:
      SO_StringCopy(pInternalName, getText(PHYSICAL_NAME));
      SO_StringCopy(pFuncName,     Q(SC_PHYSICAL));
      peType[0] = PTR_DOUBLE;
      peType[1] = PTR_DOUBLE;
      peType[2] = PTR_STRING;
      peType[3] = PTR_STRING;
      *nParamCount=4;
      break;
	 
    case FUNC_ROTATION:
      SO_StringCopy(pInternalName, getText(ROTATION_NAME));
      SO_StringCopy(pFuncName,     Q(SC_ROTATION));
      peType[0] = PTR_DOUBLE;
      peType[1] = PTR_DOUBLE;
      peType[2] = PTR_STRING;
      peType[3] = PTR_STRING;
      peType[4] = PTR_STRING;
      peType[5] = PTR_DOUBLE;
      peType[6] = PTR_DOUBLE;
      *nParamCount=7;
      break;
	 
    case FUNC_BESSELDN:
      SO_StringCopy(pInternalName, getText(BESSELDN_NAME));
      SO_StringCopy(pFuncName,     Q(SC_BESSELDN));
      peType[0] = PTR_DOUBLE;
      peType[1] = PTR_DOUBLE;
      peType[2] = PTR_STRING;
      *nParamCount=3;
      break;
	 
    case FUNC_LUNATION:
      SO_StringCopy(pInternalName, getText(LUNATION_NAME));
      SO_StringCopy(pFuncName,     Q(SC_LUNATION));
      peType[0] = PTR_DOUBLE;
      peType[1] = PTR_DOUBLE;
      *nParamCount=2;
      break;
	 
    case FUNC_MOONPHASE:
      SO_StringCopy(pInternalName, getText(MOONPHASE_NAME));
      SO_StringCopy(pFuncName,     Q(SC_MOONPHASE));
      peType[0] = PTR_DOUBLE;
      peType[1] = PTR_DOUBLE;
      peType[2] = PTR_STRING;
      peType[3] = PTR_STRING;
      *nParamCount=4;
      break;

    case FUNC_APSIDE:
      SO_StringCopy(pInternalName, getText(APSIDE_NAME));
      SO_StringCopy(pFuncName,     Q(SC_APSIDE));
      peType[0] = PTR_DOUBLE;
      peType[1] = PTR_DOUBLE;
      *nParamCount=2;
      break;
	 
    case FUNC_MOONAPSIDE:
      SO_StringCopy(pInternalName, getText(MOONAPSIDE_NAME));
      SO_StringCopy(pFuncName,     Q(SC_MOONAPSIDE));
      peType[0] = PTR_DOUBLE;
      peType[1] = PTR_DOUBLE;
      peType[2] = PTR_STRING;
      peType[3] = PTR_STRING;
      *nParamCount=4;
      break;
	 
    case FUNC_NODE:
      SO_StringCopy(pInternalName, getText(NODE_NAME));
      SO_StringCopy(pFuncName,     Q(SC_NODE));
      peType[0] = PTR_DOUBLE;
      peType[1] = PTR_DOUBLE;
      *nParamCount=2;
      break;
	 
    case FUNC_MOONNODE:
      SO_StringCopy(pInternalName, getText(MOONNODE_NAME));
      SO_StringCopy(pFuncName,     Q(SC_MOONNODE));
      peType[0] = PTR_DOUBLE;
      peType[1] = PTR_DOUBLE;
      peType[2] = PTR_STRING;
      peType[3] = PTR_STRING;
      *nParamCount=4;
      break;

    case FUNC_TROPIC:
      SO_StringCopy(pInternalName, getText(TROPIC_NAME));
      SO_StringCopy(pFuncName,     Q(SC_TROPIC));
      peType[0] = PTR_DOUBLE;
      peType[1] = PTR_DOUBLE;
      *nParamCount=2;
      break;
	 
    case FUNC_MOONMAXDECL:
      SO_StringCopy(pInternalName, getText(MOONMAXDECL_NAME));
      SO_StringCopy(pFuncName,     Q(SC_MOONMAXDECL));
      peType[0] = PTR_DOUBLE;
      peType[1] = PTR_DOUBLE;
      peType[2] = PTR_STRING;
      peType[3] = PTR_STRING;
      *nParamCount=4;
      break;
	 
    case FUNC_ECLIPSE:
      SO_StringCopy(pInternalName, getText(ECLIPSE_NAME));
      SO_StringCopy(pFuncName,     Q(SC_ECLIPSE));
      peType[0] = PTR_DOUBLE;
      peType[1] = PTR_DOUBLE;
      peType[2] = PTR_STRING;
      peType[3] = PTR_STRING;
      peType[4] = PTR_STRING;
      *nParamCount=5;
      break;
	 
    case FUNC_ECLTYPE:
      SO_StringCopy(pInternalName, getText(ECLTYPE_NAME));
      SO_StringCopy(pFuncName,     Q(SC_ECLTYPE));
      peType[0] = PTR_STRING;
      peType[1] = PTR_DOUBLE;
      *nParamCount=2;
      break;

    case FUNC_SEASON:
      SO_StringCopy(pInternalName, getText(SEASON_NAME));
      SO_StringCopy(pFuncName,     Q(SC_SEASON));
      peType[0] = PTR_DOUBLE;
      peType[1] = PTR_DOUBLE;
      peType[2] = PTR_STRING;
      peType[3] = PTR_STRING;
      *nParamCount=4;
      break;

  default:
    *nParamCount = *pFuncName = *pInternalName = 0;
    break;
  }
}


/*
 * Jede exportierte Funktion erhaelt eine kurze Erlaeuterung was sie tut,
 * welche Argumente sie erwartet usw...
 * Diese Kurzhilfe wird vom Funktionsautopiloten von StarCalc erwartet.
 *
 * Folgende Parameter muessen versorgt werden:
 *
 * USHORT *nNo:	   Nummer der Funktion zwischen 0 und nCount - 1  
 * USHORT *nParam: Nummer der Funktionsparameters
 * char   *pName:  Name des Funktionsparameters
 * char   *pDesc:  Kurze Beschreibung des Parameters
 */
SO_QUIRKSCALL( GetParameterDescription, USHORT* nNo, USHORT* nParam,
				      char* pName, char* pDesc)
{
  *pName = *pDesc = 0;

  switch (*nNo) {

#ifdef HAVE_UPDATE
    case FUNC_UPDATE:
      SO_StringCopy(pDesc, getText(UPDATE_DESC));
      break;
#endif
#ifdef INCL_SYSINFO
    case FUNC_SYSINFO:
      switch (*nParam) {
        case 0:
	  SO_StringCopy(pDesc, getText(SYSINFO_DESC));
	  break;

        case 1:
	  SO_StringCopy(pName, getText(STRING_NAME));
	  SO_StringCopy(pDesc, getText(SYSINFOS_DESC));
	  break;
      }
      break;
#endif

    case FUNC_DATE2:
      switch (*nParam) {
        case 0:
	  SO_StringCopy(pDesc, getText(DATE2_DESC));
	  break;

        case 1:
	  SO_StringCopy(pName, getText(DATE_NAME));
	  SO_StringCopy(pDesc, getText(DATE_DESC));
	  break;
      }
      break;

    case FUNC_EASTER:
      switch (*nParam) {
        case 0:
	  SO_StringCopy(pDesc, getText(EASTER_DESC));
	  break;

        case 1:
	  SO_StringCopy(pName, getText(YEAR_NAME));
	  SO_StringCopy(pDesc, getText(YEAR_DESC));
	  break;
      }
      break;

    case FUNC_DMS:
      switch (*nParam) {
        case 0:
	  SO_StringCopy(pDesc, getText(DMS_DESC));
	  break;

        case 1:
	  SO_StringCopy(pName, getText(NUMBER_NAME));
	  SO_StringCopy(pDesc, getText(NUMBER_DESC));
	  break;

        case 2:
	  SO_StringCopy(pName, getText(TAB_NAME));
	  SO_StringCopy(pDesc, getText(TAB_DESC));
	  break;

        case 3:
	  SO_StringCopy(pName, getText(UNIT_NAME));
	  SO_StringCopy(pDesc, getText(UNIT_DESC));
	  break;
      }
      break;

    case FUNC_HMS:
      switch (*nParam) {
        case 0:
	  SO_StringCopy(pDesc, getText(HMS_DESC));
	  break;

        case 1:
	  SO_StringCopy(pName, getText(NUMBER_NAME));
	  SO_StringCopy(pDesc, getText(NUMBER_DESC));
	  break;

        case 2:
	  SO_StringCopy(pName, getText(TAB_NAME));
	  SO_StringCopy(pDesc, getText(TAB_DESC));
	  break;

        case 3:
	  SO_StringCopy(pName, getText(UNIT_NAME));
	  SO_StringCopy(pDesc, getText(UNIT_DESC));
	  break;
      }
      break;

    case FUNC_DATIME:
      switch (*nParam) {
        case 0:
	  SO_StringCopy(pDesc, getText(DATIME_DESC));
	  break;

        case 1:
	  SO_StringCopy(pName, getText(NUMBER_NAME));
	  SO_StringCopy(pDesc, getText(NUMBER_DESC));
	  break;

       case 2:
	  SO_StringCopy(pName, getText(DAFMT_NAME));
	  SO_StringCopy(pDesc, getText(DAFMT_DESC));
	  break;
      }
      break;

    case FUNC_TIMEZONE:
      switch (*nParam) {
        case 0:
	  SO_StringCopy(pDesc, getText(TIMEZONE_DESC));
	  break;

        case 1:
	  SO_StringCopy(pName, getText(STRING_NAME));
	  SO_StringCopy(pDesc, getText(TZONES_DESC));
	  break;
      }
      break;

    case FUNC_POSITION:
      switch (*nParam) {
        case 0:
	  SO_StringCopy(pDesc, getText(POSITION_DESC));
	  break;

        case 1:
	  SO_StringCopy(pName, getText(DATE_NAME));
	  SO_StringCopy(pDesc, getText(DATE_DESC));
	  break;

        case 2:
	  SO_StringCopy(pName, getText(BODY_NAME));
	  SO_StringCopy(pDesc, getText(BODY_DESC));
	  break;

        case 3:
	  SO_StringCopy(pName, getText(STRING_NAME));
	  SO_StringCopy(pDesc, getText(POSITIONS_DESC));
	  break;

        case 4:
	  SO_StringCopy(pName, getText(CENTER_NAME));
	  SO_StringCopy(pDesc, getText(CENTER_DESC));
	  break;

        case 5:
	  SO_StringCopy(pName, getText(PLANE_NAME));
	  SO_StringCopy(pDesc, getText(PLANE_DESC));
	  break;

        case 6:
	  SO_StringCopy(pName, getText(TYPE_NAME));
	  SO_StringCopy(pDesc, getText(TYPE_DESC));
	  break;

        case 7:
	  SO_StringCopy(pName, getText(EQUINOX_NAME));
	  SO_StringCopy(pDesc, getText(EQUINOX_DESC));
	  break;

        case 8:
	  SO_StringCopy(pName, getText(LONGITUDE_NAME));
	  SO_StringCopy(pDesc, getText(LONGITUDE_DESC));
	  break;

        case 9:
	  SO_StringCopy(pName, getText(LATITUDE_NAME));
	  SO_StringCopy(pDesc, getText(LATITUDE_DESC));
	  break;
      }
      break;

    case FUNC_VISIBLE:
      switch (*nParam) {
        case 0:
	  SO_StringCopy(pDesc, getText(VISIBLE_DESC));
	  break;

        case 1:
	  SO_StringCopy(pName, getText(DATE_NAME));
	  SO_StringCopy(pDesc, getText(DATE_DESC));
	  break;

        case 2:
	  SO_StringCopy(pName, getText(BODY_NAME));
	  SO_StringCopy(pDesc, getText(BODY_DESC));
	  break;

        case 3:
	  SO_StringCopy(pName, getText(STRING_NAME));
	  SO_StringCopy(pDesc, getText(VISIBLES_DESC));
	  break;

        case 4:
	  SO_StringCopy(pName, getText(TIMEZONE_NAME));
	  SO_StringCopy(pDesc, getText(TZONES_DESC));
	  break;

        case 5:
	  SO_StringCopy(pName, getText(LONGITUDE_NAME));
	  SO_StringCopy(pDesc, getText(LONGITUDE_DESC));
	  break;

        case 6:
	  SO_StringCopy(pName, getText(LATITUDE_NAME));
	  SO_StringCopy(pDesc, getText(LATITUDE_DESC));
	  break;
      }
      break;

    case FUNC_PHYSICAL:
      switch (*nParam) {
        case 0:
	  SO_StringCopy(pDesc, getText(PHYSICAL_DESC));
	  break;

        case 1:
	  SO_StringCopy(pName, getText(DATE_NAME));
	  SO_StringCopy(pDesc, getText(DATE_DESC));
	  break;

        case 2:
	  SO_StringCopy(pName, getText(BODY_NAME));
	  SO_StringCopy(pDesc, getText(BODY_DESC));
	  break;

        case 3:
	  SO_StringCopy(pName, getText(STRING_NAME));
	  SO_StringCopy(pDesc, getText(PHYSICALS_DESC));
	  break;
      }
      break;

    case FUNC_ROTATION:
      switch (*nParam) {
        case 0:
	  SO_StringCopy(pDesc, getText(ROTATION_DESC));
	  break;

        case 1:
	  SO_StringCopy(pName, getText(DATE_NAME));
	  SO_StringCopy(pDesc, getText(DATE_DESC));
	  break;

        case 2:
	  SO_StringCopy(pName, getText(BODY_NAME));
	  SO_StringCopy(pDesc, getText(BODY_DESC));
	  break;

        case 3:
	  SO_StringCopy(pName, getText(CENTER_NAME));
	  SO_StringCopy(pDesc, getText(CENTER_DESC));
	  break;

        case 4:
	  SO_StringCopy(pName, getText(STRING_NAME));
	  SO_StringCopy(pDesc, getText(ROTATIONS_DESC));
	  break;

        case 5:
	  SO_StringCopy(pName, getText(LONGITUDE_NAME));
	  SO_StringCopy(pDesc, getText(LONGITUDE_DESC));
	  break;

        case 6:
	  SO_StringCopy(pName, getText(LATITUDE_NAME));
	  SO_StringCopy(pDesc, getText(LATITUDE_DESC));
	  break;
      }
      break;

    case FUNC_BESSELDN:
      switch (*nParam) {
        case 0:
	  SO_StringCopy(pDesc, getText(BESSELDN_DESC));
	  break;

        case 1:
	  SO_StringCopy(pName, getText(DATE_NAME));
	  SO_StringCopy(pDesc, getText(DATE_DESC));
	  break;

        case 2:
	  SO_StringCopy(pName, getText(STRING_NAME));
	  SO_StringCopy(pDesc, getText(BESSELDNS_DESC));
	  break;
      }
      break;

    case FUNC_LUNATION:
      switch (*nParam) {
        case 0:
	  SO_StringCopy(pDesc, getText(LUNATION_DESC));
	  break;

        case 1:
	  SO_StringCopy(pName, getText(DATE_NAME));
	  SO_StringCopy(pDesc, getText(DATE_DESC));
	  break;
      }
      break;

    case FUNC_MOONPHASE:
      switch (*nParam) {
        case 0:
	  SO_StringCopy(pDesc, getText(MOONPHASE_DESC));
	  break;

        case 1:
	  SO_StringCopy(pName, getText(LUNATION_NAME));
	  SO_StringCopy(pDesc, getText(LUNATION_DESC));
	  break;

        case 2:
	  SO_StringCopy(pName, getText(STRING_NAME));
	  SO_StringCopy(pDesc, getText(LUNARPHASES_DESC));
	  break;

        case 3:
	  SO_StringCopy(pName, getText(TIMEZONE_NAME));
	  SO_StringCopy(pDesc, getText(TZONES_DESC));
	  break;
      }
      break;

    case FUNC_APSIDE:
      switch (*nParam) {
        case 0:
	  SO_StringCopy(pDesc, getText(APSIDE_DESC));
	  break;

        case 1:
	  SO_StringCopy(pName, getText(DATE_NAME));
	  SO_StringCopy(pDesc, getText(DATE_DESC));
	  break;
      }
      break;

    case FUNC_MOONAPSIDE:
      switch (*nParam) {
        case 0:
	  SO_StringCopy(pDesc, getText(MOONAPSIDE_DESC));
	  break;

        case 1:
	  SO_StringCopy(pName, getText(APSIDE_NAME));
	  SO_StringCopy(pDesc, getText(APSIDE_DESC));
	  break;

        case 2:
	  SO_StringCopy(pName, getText(STRING_NAME));
	  SO_StringCopy(pDesc, getText(LUNARAPSIDES_DESC));
	  break;

        case 3:
	  SO_StringCopy(pName, getText(TIMEZONE_NAME));
	  SO_StringCopy(pDesc, getText(TZONES_DESC));
	  break;
      }
      break;

    case FUNC_NODE:
      switch (*nParam) {
        case 0:
	  SO_StringCopy(pDesc, getText(NODE_DESC));
	  break;

        case 1:
	  SO_StringCopy(pName, getText(DATE_NAME));
	  SO_StringCopy(pDesc, getText(DATE_DESC));
	  break;
      }
      break;

    case FUNC_MOONNODE:
      switch (*nParam) {
        case 0:
	  SO_StringCopy(pDesc, getText(MOONNODE_DESC));
	  break;

        case 1:
	  SO_StringCopy(pName, getText(NODE_NAME));
	  SO_StringCopy(pDesc, getText(NODE_DESC));
	  break;

        case 2:
	  SO_StringCopy(pName, getText(STRING_NAME));
	  SO_StringCopy(pDesc, getText(LUNARNODES_DESC));
	  break;

        case 3:
	  SO_StringCopy(pName, getText(TIMEZONE_NAME));
	  SO_StringCopy(pDesc, getText(TZONES_DESC));
	  break;
      }
      break;

    case FUNC_TROPIC:
      switch (*nParam) {
        case 0:
	  SO_StringCopy(pDesc, getText(TROPIC_DESC));
	  break;

        case 1:
	  SO_StringCopy(pName, getText(DATE_NAME));
	  SO_StringCopy(pDesc, getText(DATE_DESC));
	  break;
      }
      break;

    case FUNC_MOONMAXDECL:
      switch (*nParam) {
        case 0:
	  SO_StringCopy(pDesc, getText(MOONMAXDECL_DESC));
	  break;

        case 1:
	  SO_StringCopy(pName, getText(TROPIC_NAME));
	  SO_StringCopy(pDesc, getText(TROPIC_DESC));
	  break;

        case 2:
	  SO_StringCopy(pName, getText(STRING_NAME));
	  SO_StringCopy(pDesc, getText(LUNARMAXDECLS_DESC));
	  break;

        case 3:
	  SO_StringCopy(pName, getText(TIMEZONE_NAME));
	  SO_StringCopy(pDesc, getText(TZONES_DESC));
	  break;
      }
      break;

    case FUNC_ECLIPSE:
      switch (*nParam) {
        case 0:
	  SO_StringCopy(pDesc, getText(ECLIPSE_DESC));
	  break;

        case 1:
	  SO_StringCopy(pName, getText(LUNATION_NAME));
	  SO_StringCopy(pDesc, getText(LUNATION_DESC));
	  break;

        case 2:
	  SO_StringCopy(pName, getText(STRING_NAME));
	  SO_StringCopy(pDesc, getText(ECLIPSES_DESC));
	  break;

        case 3:
	  SO_StringCopy(pName, getText(STRING_NAME));
	  SO_StringCopy(pDesc, getText(ECLVALUE_DESC));
	  break;

        case 4:
	  SO_StringCopy(pName, getText(TIMEZONE_NAME));
	  SO_StringCopy(pDesc, getText(TZONES_DESC));
	  break;
      }
      break;

    case FUNC_ECLTYPE:
      switch (*nParam) {
        case 0:
	  SO_StringCopy(pDesc, getText(ECLTYPE_DESC));
	  break;

        case 1:
	  SO_StringCopy(pName, getText(NUMBER_NAME));
	  SO_StringCopy(pDesc, getText(ECLTYPES_DESC));
	  break;
      }
      break;

    case FUNC_SEASON:
      switch (*nParam) {
        case 0:
	  SO_StringCopy(pDesc, getText(SEASON_DESC));
	  break;

        case 1:
	  SO_StringCopy(pName, getText(YEAR_NAME));
	  SO_StringCopy(pDesc, getText(YEAR_DESC));
	  break;

        case 2:
	  SO_StringCopy(pName, getText(STRING_NAME));
	  SO_StringCopy(pDesc, getText(SEASONS_DESC));
	  break;

        case 3:
	  SO_StringCopy(pName, getText(TIMEZONE_NAME));
	  SO_StringCopy(pDesc, getText(TZONES_DESC));
	  break;
      }
      break;
  }
}


